; Fahrtregler der Luebeck II
; Misst Laenge des PWM Pulses und gibt einen entsprechenden 8-Bit Fahrtwert sowie Gate-Signale 
; fuer die HS und LS FET aus
; Bei Durchgang durch Fahrtstufe 0 wird ein delay eingefuegt
; PWM-Ein    = RA4
; Fahrtwert = RB0-RB7
; HS-rechts = RA3
; HS-links = RA2
; LS-links = RA1
; LS-rechts = RA0
; HS_xx: High bedeutet FET schaltet durch (High wird durch Treiber auf Low invertiert, Low auf P-FET = ein)
;	Low bedeutet FET sperrt (Gate ist ueber Collector-Widerstand der Treiberschaltung auf High)
; LS_xx: Low bedeutet PWM ist auf FET geschaltet (Switch-Schaltung ist durchlaessig bei low auf Basis)
;	High bedeutet PWM ist fuer den FET gesperrt (Gate wird durch Transistor der Switch-Schaltung auf low gezogen)

;MCU wird mit externem RC geclockt -> ca. 4,869564 MHz (=4,xMHz)
; Interner Timer wird geclockt mit 4,xMhz/4 = 1,217391MHz (0,821429us), Prescaler ist auf 1:8
; -> resolution ist 0,006571ms, maximaler Zaehlerstand ist 1,67571ms
;Bei Beginn eines Pulses wird zuerst 0,7ms gewartet und dann erst die Zeitmessung ueber timer gestartet
;-> ermoeglicht 0,006571ms Aufloesung ohne Timer-Ueberlauf auch bei Pulslaengen >2ms
;Damit ergeben sich folgende Pulslaengen: 
;Voll vorwaerts: (1,0ms-0,7ms)/0,006571ms=46, Mittelstellung: (1,5ms-0,7ms)/0,006571ms=122, Voll rueckwaerts: (2,0ms-0,7ms)/0,006571ms=198
;PWM Ausgang wird 8 Bit breit betrieben -> maximale PWM ist 255/256 duty cycle
;Es bestuende auch die Moeglichkeit den PWM Wert nur 7 Bit brei auszugeben (100% duty cycle) 
;dies wurde aber die PWM Frequenz verdoppeln, d.h. FETs waeren schwerer zu treiben

;;Config-register sollte bei Programmierung auf 11111111110011 (3FF3) stehen
;;(cp=1: code protection off, wdte=0: power up timer on, wdt=0: kein watchdog timer, fosc=11: RC oszillator) 
;RC Oszillator wurde gewaehlt da dieser bei ClkIn=4MHz auf ClkOut 1MHz statt 4MHz (XT) ausgibt (reduziert die PWM Frequenz)

	processor  16F84A
	radix  DEC
	__CONFIG 0x3FF3

INDF    EQU   	0x00
TMR0    EQU   	0x01
STATUS  EQU 	0x03
FSR     EQU 	0x04
PORTA   EQU 	0x05
PORTB	EQU		0x06
INTCON  EQU 	0x0B
TRISA   EQU 	0x85
TRISB	EQU 	0x86

_C      EQU   	0		;Carry-Bit
_Z      EQU   	2		;Zero-Bit
_B		EQU		5		;Bank-Bit
_W		EQU		0		;Destination W bei z.B. INCF
_R		EQU		1		;Destination selbes Register bei z.B. INCF 
PWM     EQU   	4		;PWM Eingang
HSL     EQU   	2		;High-side links
HSR     EQU   	3		;High-side rechts
LSL     EQU   	1		;Low-side links
LSR     EQU   	0		;Low-side rechts
				
VWmin	EQU   	118		;Pulsdauer, bei der minimale Vorwaertsfahrt anstehen soll (Mittelstellung - 4)
VWmax	EQU   	46		;Pulsdauer, bei der maximale Vorwaertsfahrt anstehen soll (Hebel voll oben)
RWmin	EQU   	126		;Pulsdauer, bei der minimale Rueckwaertsfahrt anstehen soll (Mittelstellung +4)
RWmax	EQU   	198		;Pulsdauer, bei der maximale Rueckwaertsfahrt anstehen soll (Hebel voll unten)
FWmin	EQU   	100		;Fahrtwert fuer Minimalfahrt (Motor laeuft an bei ca 5V -> 64 bei 7-bit PWM, 128 bei 8-bit PWM)
FWmax	EQU   	255		;Fahrtwert fuer Maximalfahrt (maximale PWM -> 128 bei 7-bit, 255 bei 8-bit PWM)
Skal	EQU   	64		;Skalierungsfaktor (64 bei 8-bit PWM, 128 bei 7-bit PWM)
;!!!!!!Folgende 8 Werte muesen aus den obigen haendisch errechnet werden!!!!!
;!!!!!!Geht leider nicht automatisch: Assembler Arithmetik rundet jede Division sofort
;Vorwaertsfahrt: VWStg=(FWmax-FWmin)/(VWmin-VWmax)*Skal - ergibt ein 16-Bit Wort (113 = 0 und 113)
;Vorwaertsfahrt: VWOff=[((FWmax-FWmin)/(VWmin-VWmax)*VWmax)+FWmax]*Skal - ergibt ein 16-bit Wort (43518 = 169 und 254)
VWStgL	EQU   	138		;Steigung Vorwaetrs Lowword
VWStgH	EQU   	0		;Steigung Vorwaetrs Highword
VWOffL	EQU   	130	   	;Offset Vorwaerts Lowword
VWOffH	EQU   	88		;Offset Vorwaerts Highword
;Rueckwaertsfahrt: RWStg=(FWmax-FWmin)/(RWmax-RWmin)*Skal - ergibt ein 16-Bit Wort (232 = 0 und 232)
;Rueckwaertsfahrt: RWOff=[((FWmax-FWmin)/(RWmax-RWmin)*RWmin)-FWmin]*Skal - ergibt ein 16-bit Wort (19908 = 77 und 196)
RWStgL	EQU   	138		;Steigung Rueckwaerts Lowword
RWStgH	EQU   	0		;Steigung Rueckwaerts Highword
RWOffL	EQU   	208		;Offset der Gerade bei Rueckwaertsfahrt - Lowword
RWOffH	EQU   	42		;Offset der Gerade bei Rueckwaertsfahrt - Highword

MEM		EQU		0x0C
puls	EQU		MEM+0	;Speichert die Dauer eines PWM Pulses
flags	EQU		MEM+1	;Bit0: Letzter Wert war schon 0, Bit1: Delay einleiten 
x		EQU		MEM+2	;Allgemeiner Zaehler
fw		EQU		MEM+3	;Zwischenspeicher fuer Fahrtwert
el		EQU		MEM+4	;Ergebnis - Lowword
eh		EQU		MEM+5	;Ergebnis - Highword
o1l		EQU		MEM+6	;Operand1 - Lowword
o1h		EQU		MEM+7	;Operand1 - Highword
o2l		EQU		MEM+8	;Operand2 - Lowword
o2h		EQU		MEM+9	;Operand3 - Highword


		GOTO	main

;***********************************************************
DelayL						;Wartephase bei Nullfahrt-Eintritt
		MOVLW  	.255		;zusammen mit dem aufrufenden CALL
		MOVWF  	x			;werden hier 2045 Zyklen gewartet
w1		NOP					;2+2+(254*8)+7+2 (Call + 2xMOV + 16xSchleife +1xSchleifenaustritt)
		NOP
		NOP
		NOP
		NOP
		DECFSZ 	x,1
		GOTO  	w1
		RETLW 	0

;**************************************************************
Delay5						;Wartephase 0,7ms fuer Pulsmessung
		MOVLW  	.105			;zusammen mit dem aufrufenden CALL
		MOVWF  	x			;werden hier 852 Zyklen (*0,821429us = 0,7ms) gewartet
w2		NOP					;2+2+(104*8)+7+7+2 (Call + 2xMOV + 104xSchleife +1xSchleifenaustritt)
		NOP
		NOP
		NOP
		NOP
		DECFSZ 	x,1
		GOTO  	w2
		NOP
		NOP
		NOP
		NOP
		NOP
		NOP
		NOP
		RETLW 	0

;******* Division *******************************************************
Div		;Shifted eh/el nach rechts und /64 bzw /128 zu realisieren. Rechts faellt's raus, links wird 0 aufgefuellt
;		MOVLW	.7			;Sieben Iterationen: Division /128 (bei Skal =128)
		MOVLW	.6			;Sechs Iterationen: Division /64 (bei Skal=64)
		MOVWF	x
Dloop	BCF		STATUS,_C	;mit 0 auffuellen
		RRF		eh,_R		;eh Rechtsshiften unterstes Bit nach carry	
		RRF		el,_R		;el Rechtsshiften, oben wird mit carry aufgefuellt
		DECFSZ	x,_R
		GOTO	Dloop
		RETLW 	0


;******* Addit *******************************************************
Addit	;Addiert E=E+O2, E und O2 sind jeweils 16 Bit breit
		MOVF	o2h,_W		;E aendert sich, o2h/o2l bleiben unveraendert
		ADDWF	eh,_W		;W=eh+o2h
		MOVWF	eh			;eh=W
		MOVF	o2l,_W
		ADDWF	el,_W		;W=el+o2l
		MOVWF	el			;el=W
		BTFSC	STATUS,_C	;Pruefe, ob Ueberlauf
		INCF	eh,_R		;Ueberlauf: eh erhoehen
		RETLW	0		


;******* Subtr *******************************************************
Subtr	;Subtrahiert 16bit o2 von 16bit o1 zu 16bit e.
		MOVF	o2h,_W		;o2l/o2h, o1l/o1h bleiben unveraendert, e hat danach Ergebnis
		SUBWF	o1h,_W		;W=o1h-o2h
		MOVWF	eh			;eh=W
		MOVF	o2l,_W
		SUBWF	o1l,_W		;W=o1l-o2l
		MOVWF	el			;el=W
		BTFSS	STATUS,_C	;Pruefe, ob Ueberlauf
		DECF	eh,_R		;Ueberlauf: eh erniedrigen
		RETLW	0	


;******* Multip *******************************************************
Multip	;Multipliziert 16bit mit 8bit zu 16bit Ergebnis. el/eh ist Ergebnis, o1l ist 8bit Wert, o2l/o2h ist 16bit Wert
		;Es muss sichergestellt werden, dass Operanden so sind, dass kein Ueberlauf stattfindet!
		CLRF	el			;Alle Werte (o1l, o2l, o2h, el, eh aendern sich durch die Funktion)
		CLRF	eh			;Clear Ergebnis
		MOVLW   .8			;8-fach Schleife
		MOVWF   x
Mloop	RRF		o1l,_R		;Right shift 8-bit-Wert nach carry
		BTFSC	STATUS, _C	;Check geshiftetes Bit
		CALL	Addit		;Wenn 1: E=E+O2 
		BCF		STATUS,_C	;Es wird mit 0 aufgefuellt
		RLF		o2l,_R		;Lowbyte shiften oberster Wert nach carry
		RLF		o2h,_R		;Highbyte shiften und mit Carry fuellen
		DECFSZ	x,_R		;Schleifenzaehler -1, Zaehler=0?
		GOTO	Mloop		;Nein: Neue Schleife
		RETLW 	0



main	
		BCF   	STATUS,_B	;auf Bank 0 schalten
		CLRF  	INTCON		;0Bh: INTCON: Alle interrupts disabled
		MOVLW 	0x82				
		OPTION				;Option: pullup disabled, timer input intern, prescaler assigned to timer, prescale by 1:8	
		BSF   	STATUS,_B	;auf Bank 1 schalten
		MOVLW 	0x10
		MOVWF 	TRISA		;85h: TRISA: RA4 ist input (PWM), RA0-3 sind Output (Gate control)
		CLRF 	TRISB		;86h: TRISB: alle RBx sind output (fuer Fahrtwert)
		BCF   	STATUS,_B	;auf bank 0 schalten
		CLRF  	PORTB		;Fahrtwert=0
		MOVLW	0x03		;Alle HS/LS schliessen
		MOVWF	PORTA		;Grund: Bei 9V power-up sind HS geschlossen (pull-up)/LS evtl. offen (PWM floating). Gleich auf 1111 zu gehen koennte kurzen Kurzschluss machen
		MOVLW	0xFF		;Beide HS auf, LS bleiben geschlossen
		MOVWF	PORTA
	
m001	BTFSS 	PORTA,PWM	;positiver Puls ?
		GOTO  	m003		;nein, warte auf Puls
m002	BTFSC 	PORTA,PWM	;ja, Ende des Pulses erreicht?
		GOTO  	m002		;nein, weiter warten
		;Definierter Zustand erreicht, PWM befindet sich am Anfang der 22ms langen Pulsruhezeit
m003	BTFSS 	PORTA,PWM	;Warte auf Beginn eines Pulses
		GOTO  	m003	
		;Der positive Puls hat begonnen
		CALL	Delay5		;0,5ms warten
		CLRF  	TMR0		;Zeitmessung starten
		BTFSS	PORTA,PWM	;Ist Puls nach den 0,5ms noch aktiv?
		GOTO	m003		;Nein -> war wohl Fehlpuls (<0,5ms) -> auf naechsten Puls warten			
m004	BTFSC 	PORTA,PWM	;Ja: Warte auf Ende des Pulses
		GOTO  	m004
		;Der Puls ist zu Ende
		MOVF  	TMR0,W		;Pulsdauer in Variable festhalten
		MOVWF 	puls
		
		;Pulsdauer bewerten
		MOVLW	VWmin
		ADDLW	1			;VWmin+1 = unterster Wert fuer Nullfahrt (Nullmin) nach W
		SUBWF	puls,_W		;W=Puls-Nullmin
		BTFSS	STATUS,_C	;Carry=0:Ueberlauf (Puls<Nullmin)?
		GOTO	F_Vorw		;ja: Vorwaertsfahrt
		MOVLW	RWmin
		MOVWF	x
		DECF	x,_W		;RWmin-1 = oberster Wert fuer Nullfahrt (Nullmax) nach W
		SUBWF	puls,_W		;W=Puls-Nullmax
		BTFSS	STATUS,_C	;Carry=0:Ueberlauf (Puls<Nullmax)?
		GOTO	F_Null		;ja: Nullfahrt	
F_Rueck	MOVLW	RWmax		;nein: -> Puls>Nullmax: Rueckwaertsfahrt	
		SUBWF	puls,_W		;W=Puls-RWmax
		BTFSS	STATUS,_C	;Carry=0:Ueberlauf (Puls<RWmax)?
		GOTO	RWLin		;ja: Puls innerhalb des linearen Bereichs -> Gerade berechnen
		MOVLW	FWmax		;nein: Puls laenger als max-wert -> fw ist maxwert
		MOVWF	fw
		GOTO	RWein		;Steuersignale einstellen
RWLin	;linearen Fahrtwert berechnen: FW = (RWStg * Puls - RWOff) / Skal
		MOVLW	RWStgH		;o2h/o2l mit Steigungswert fuellen
		MOVWF	o2h
		MOVLW	RWStgL
		MOVWF	o2l
		MOVF	puls,_W		;o1l mit Puls fuellen
		MOVWF	o1l
		CALL	Multip		;E=RWStg*Puls
		MOVF	el,_W			;o1h/o1l mit Ergebnis fuellen
		MOVWF	o1l
		MOVF	eh,_W
		MOVWF	o1h
		MOVLW	RWOffH		;o2h/o2l mit RWOff befuellen
		MOVWF	o2h
		MOVLW	RWOffL
		MOVWF	o2l
		CALL	Subtr		;E=(RWStg*Puls)-RWOff
		CALL	Div			;E=[(RWStg*Puls)-RWOff]/Skal (es sollte jetzt nur el Inhalt haben)
		MOVF	el,_W
		MOVWF	fw			;fw=el
RWein		MOVLW	0x09		;Gatedriver einstellen (HSR,HSL,LSL,LSR)=1001
		MOVWF	PORTA		;HSR schaltet durch, PWM auf LSL
		CALL	Delay5		;Kleine Pause, damit 2. HS auch wirklich schon dicht (wenn aus Nullfahrt kommend), bevor LS per PWM aufgeht
		MOVF	fw,_W			
		MOVWF	PORTB		;PWM einstellen
		BCF		flags,0		;Letzter-Wert-war-null-Flag loeschen
		GOTO	m003		;Weiter zum naechsten Puls
F_Vorw	MOVLW	VWmax	
		SUBWF	puls,_W		;W=Puls-VWmax
		BTFSC	STATUS,_C	;Carry=1: kein Ueberlauf (Puls>VWmax)?
		GOTO	VWLin		;ja: Puls innerhalb des linearen Bereichs -> Gerade berechnen
		MOVLW	FWmax		;nein: Puls kuerzer als max-wert -> fw ist maxwert
		MOVWF	fw
		GOTO	VWein		;Steuersignale einstellen
VWLin	;linearen Fahrtwert berechnen: FW = (VWOff - VWStg * Puls) / Skal
		MOVLW	VWStgH		;o2h/o2l mit Steigungswert fuellen
		MOVWF	o2h
		MOVLW	VWStgL
		MOVWF	o2l
		MOVF	puls,_W		;o1l mit Puls fuellen
		MOVWF	o1l
		CALL	Multip		;E=VWStg*Puls
		MOVF	el, _W			;o2h/o2l mit Ergebnis fuellen
		MOVWF	o2l
		MOVF	eh, _W
		MOVWF	o2h
		MOVLW	VWOffH		;o1h/o1l mit VWOff befuellen
		MOVWF	o1h
		MOVLW	VWOffL
		MOVWF	o1l
		CALL	Subtr		;E=VWOff-(VWStg*Puls)
		CALL	Div			;E=[VWOff-(VWStg*Puls)]/Skal (es sollte jetzt nur el Inhalt haben)
		MOVF	el,_W
		MOVWF	fw			;fw=el
VWein		MOVLW	0x06		;Gatedriver einstellen (HSR,HSL,LSL,LSR)=0110
		MOVWF	PORTA		;HSL schaltet durch, PWM auf LSR
		CALL	Delay5		;Kleine Pause, damit 2. HS auch wirklich schon dicht (wenn aus Nullfahrt kommend), bevor LS per PWM aufgeht
		MOVF	fw,_W
		MOVWF	PORTB		;PWM einstellen
		BCF		flags,0		;Letzter-Wert-war-null-Flag loeschen
		GOTO	m003		;Weiter zum naechsten Puls
F_Null	BTFSS	flags,0		;War letzter Wert schon mal 0?
		BSF		flags,1		;Nein - d.h. Eintritt in Nullfahrt -> delay-flag=1
		BSF		flags,0		;Letzter-Wert-war-null-Flag setzen
		CLRF	PORTB		;PWM auf Null setzen
		CALL	Delay5		;Kleine Pause, damit LS auch wirklich beide dicht (wenn aus Fahrt kommend) bevor HS auf gehen
		MOVLW	0x0F		;Gatedriver einstellen (HSR,HSL,LSL,LSR)=1111		
		MOVWF	PORTA		;Beide HS schalten durch, PWM fuer beide LS gesperrt
		BTFSC	flags,1		;Delay-Flag gesetzt?
		CALL	DelayL		;Ja, Langer Delay
		BCF		flags,1		;Delay-Flag loeschen
		GOTO  	m001		;Naechster Puls (man muss wieder nach m001, weil nach DelayL unbekannter Einstiegswert

	END
